/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdbool.h>
#include <rte_mbuf.h>
#include <rte_ethdev.h>
#include <rte_mbuf_core.h>

#include <doca_dpi_grpc_client.h>
#include <doca_log.h>
#include <doca_argp.h>

#include <dpdk_utils.h>

#include "dpi_common.h"

DOCA_LOG_REGISTER(GRPC_DPI_SCAN::MAIN);

/* Sample's Logic */
int grpc_dpi_scan(const char *sig_file, const struct doca_dpi_config_t *config, struct doca_dpi_grpc_generic_packet *packet,
		  struct doca_dpi_parsing_info *parsing_info, uint32_t *payload_offset);

/*
 * Sample main function
 *
 * @argc [in]: command line arguments size
 * @argv [in]: array of command line arguments
 * @return: EXIT_SUCCESS on success and EXIT_FAILURE otherwise
 */
int
main(int argc, char **argv)
{
	int ret;
	doca_error_t result;
	struct dpi_scan_config cfg = { .is_grpc_mode = 1 };
	struct application_dpdk_config dpdk_config = {
		.port_config.nb_ports = 1,
		.port_config.nb_queues = 1,
		.reserve_main_thread = true,
	};
	bool packet_inspected = false;
	int packet_number;
	uint8_t port = 0;
	uint16_t nb_rx;
	uint16_t queue_id = 0;
	uint16_t burst_size = 8;
	uint32_t payload_offset = 0;
	struct rte_mbuf *dpdk_packets[burst_size];
	struct doca_dpi_grpc_generic_packet packet;
	uint8_t buffer[0xffff];
	struct doca_dpi_config_t config = {
		.max_packets_per_queue = 1024,
		.max_sig_match_len = 500,
		.nb_queues = 1,
	};
	struct doca_dpi_parsing_info parsing_info = {
		/* Hardcoded values */
		.ethertype = rte_cpu_to_be_16(0x0800),
		.l4_protocol = IPPROTO_UDP,
		.l4_dport = 776,
		.l4_sport = 775,
		.dst_ip.ipv4.s_addr = RTE_IPV4(127, 0, 0, 1),
	};

	result = doca_argp_init("grpc_dpi_scan", &cfg);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to init ARGP resources: %s", doca_get_error_string(result));
		return EXIT_FAILURE;
	}
	doca_argp_set_dpdk_program(dpdk_init);
	doca_argp_set_grpc_program();
	result = register_dpi_scan_params();
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to parse sample input: %s", doca_get_error_string(result));
		doca_argp_destroy();
		return EXIT_FAILURE;
	}

	result = doca_argp_start(argc, argv);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to parse sample input: %s", doca_get_error_string(result));
		doca_argp_destroy();
		return EXIT_FAILURE;
	}

	result = doca_argp_get_grpc_addr(&config.server_address);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to parse IP input: %s", doca_get_error_string(result));
		doca_argp_destroy();
		return EXIT_FAILURE;
	}

	dpdk_queues_and_ports_init(&dpdk_config);

	packet.segment = buffer;
	while (!packet_inspected) {
		nb_rx = rte_eth_rx_burst(port, queue_id, dpdk_packets, burst_size);
		if (nb_rx != 0) {
			for (packet_number = 0; packet_number < nb_rx; packet_number++) {
				packet.seg_len = rte_pktmbuf_data_len(dpdk_packets[packet_number]);
				memcpy(buffer, rte_pktmbuf_mtod(dpdk_packets[packet_number], void *), packet.seg_len);
				ret = grpc_dpi_scan(cfg.sig_file_path, &config, &packet, &parsing_info,
						    &payload_offset);
				if (ret != 0) {
					DOCA_LOG_ERR("Scan of packet %d failed! Exiting", packet_number);
					dpdk_queues_and_ports_fini(&dpdk_config);
					dpdk_fini();
					doca_argp_destroy();
					return EXIT_FAILURE;
				}
				rte_pktmbuf_free(dpdk_packets[packet_number]);
			}
			packet_inspected = true;
		}
	}

	dpdk_queues_and_ports_fini(&dpdk_config);
	dpdk_fini();
	doca_argp_destroy();

	return EXIT_SUCCESS;
}
